static const struct setting_history_rename settings_history_pigeonhole_renames[] = {
};
static const struct setting_history_default settings_history_pigeonhole_defaults[] = {
  { "service/managesieve/service_extra_groups", "$SET:default_internal_group", "3.2.0" },
};
