#ifndef EXT_REGEX_COMMON_H
#define EXT_REGEX_COMMON_H

/*
 * Extension
 */

extern const struct sieve_extension_def regex_extension;

/*
 * Operand
 */

extern const struct sieve_operand_def regex_match_type_operand;

/*
 * Match type
 */

extern const struct sieve_match_type_def regex_match_type;

#endif


