#ifndef PIGEONHOLE_SETTINGS_H
#define PIGEONHOLE_SETTINGS_H

extern const char *pigeonhole_settings_version;
extern const char *pigeonhole_settings_doveconf_banner;

extern const struct setting_parser_info *pigeonhole_settings_set_infos[];

#endif
