#ifndef EXT_VARIABLES_ARGUMENTS_H
#define EXT_VARIABLES_ARGUMENTS_H

#include "sieve-common.h"

/*
 * Variable argument
 */

extern const struct sieve_argument_def variable_argument;

/*
 * Match value argument
 */

extern const struct sieve_argument_def match_value_argument;

/*
 * Variable string argument
 */

extern const struct sieve_argument_def variable_string_argument;

#endif
